﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Points;

namespace Grids { //todo: xml comments, unit testing? same question for point & rectangle.
	public static class ExtGet {
		public static IEnumerable<U> Get<T, U>(this IEnumerable<T> source, Func<T, IEnumerable<U>> selector) {
			return source.SelectMany(selector);
		}
		public static IEnumerable<U> Get<T, U>(this IEnumerable<T> source, Func<T, U> selector) { //todo, move this somewhere cool
			return source.Select(selector);
		}
	}
	public interface ReadonlyGrid<T> : IEnumerable<T>, Rectangular where T : class {
		int Rows { get; }
		int Cols { get; }
		T this[point p] { get; }
		T this[Positioned positioned] { get; }
		IEnumerable<T> this[IEnumerable<point> positions] { get; }
		IEnumerable<T> this[IEnumerable<Positioned> positionedElements] { get; }
	}
	public class Grid<T> : ReadonlyGrid<T>, IEnumerable<T>, Rectangular where T : class {
		protected T[,] objs;
		protected HashSet<T> denseObjs;
		public int Rows { get; protected set; }
		public int Cols { get; protected set; }
		public rectangle rect => new rectangle(point.Zero, new point(Rows, Cols));
		public T this[point p] {
			get {
				if(rect.Contains(p)) return objs[p.row,p.col];
				else return null;
			}
			protected set {
				objs[p.row,p.col] = value;
			}
		}
		public T this[Positioned positioned] => this[positioned.p];
		public IEnumerable<T> this[IEnumerable<point> positions] {
			get {
				HashSet<point> returned = new HashSet<point>();
				foreach(point p in positions) {
					T t = this[p];
					if(t != null && returned.Add(p)) {
						yield return t;
					}
				}
			}
		}
		public IEnumerable<T> this[IEnumerable<Positioned> positionedElements] {
			get {
				HashSet<point> returned = new HashSet<point>();
				foreach(Positioned psd in positionedElements) {
					T t = this[psd.p];
					if(t != null && returned.Add(psd.p)) {
						yield return t;
					}
				}
			}
		}
		public bool Move(point source,point destination) {
			// Can only move an existing item, and only to a point in-bounds and empty.
			if(this[source] != null && rect.Contains(destination) && this[destination] == null) {
				this[destination] = this[source];
				this[source] = null;
				return true;
			}
			return false;
		}
		public bool Swap(point source1, point source2) {
			// Can only swap two existing items.
			if(this[source1] != null && this[source2] != null) {
				T t = this[source1];
				this[source1] = this[source2];
				this[source2] = t;
				return true;
			}
			return false;
		}
		public bool Add(T element,point destination) {
			// Can only add a non-null item that isn't already on the grid, and only to a point in-bounds and empty.
			if(element != null && rect.Contains(destination) && this[destination] == null && denseObjs.Add(element)) {
				this[destination] = element;
				return true;
			}
			return false;
		}
		public bool Remove(point source) {
			// Can only remove an existing item.
			if(this[source] != null && denseObjs.Remove(this[source])) {
				this[source] = null;
				return true;
			}
			return false;
		}
		IEnumerator IEnumerable.GetEnumerator() => this.GetEnumerator();
		public IEnumerator<T> GetEnumerator() { return denseObjs.GetEnumerator(); }
		public Grid(int rows, int cols) {
			Rows = rows;
			Cols = cols;
			objs = new T[rows, cols];
			denseObjs = new HashSet<T>();
		}
	}
	public interface ReadonlyMultiGrid<T> : IEnumerable<T>, Rectangular where T : class {
		int Rows { get; }
		int Cols { get; }
		IEnumerable<T> this[point p] { get; }
		IEnumerable<T> this[Positioned positioned] { get; }
		IEnumerable<T> this[IEnumerable<point> positions] { get; }
		IEnumerable<T> this[IEnumerable<Positioned> positionedElements] { get; }
	}
	public class MultiGrid<T> : ReadonlyMultiGrid<T>, IEnumerable<T>, Rectangular where T : class {
		protected List<T>[,] objs;
		protected HashSet<T> denseObjs;
		public int Rows { get; protected set; }
		public int Cols { get; protected set; }
		public rectangle rect => new rectangle(point.Zero, new point(Rows, Cols));
		public IEnumerable<T> this[point p] {
			get {
				if(rect.Contains(p) && objs[p.row,p.col] != null) return objs[p.row, p.col];
				else return Enumerable.Empty<T>();
			}
		}
		public IEnumerable<T> this[Positioned positioned] => this[positioned.p];
		public IEnumerable<T> this[IEnumerable<point> positions] {
			get {
				HashSet<point> returned = new HashSet<point>();
				foreach(point p in positions) {
					if(rect.Contains(p) && objs[p.row, p.col] != null && returned.Add(p)) {
						foreach(T t in objs[p.row,p.col]) {
							yield return t;
						}
					}
				}
			}
		}
		public IEnumerable<T> this[IEnumerable<Positioned> positionedElements] {
			get {
				HashSet<point> returned = new HashSet<point>();
				foreach(Positioned psd in positionedElements) {
					if(rect.Contains(psd.p) && objs[psd.p.row, psd.p.col] != null && returned.Add(psd.p)) {
						foreach(T t in objs[psd.p.row, psd.p.col]) {
							yield return t;
						}
					}
				}
			}
		}
		public bool Move(T element, point source, point destination) {
			// Can only move an existing item, and only to a point in-bounds.
			if(element != null && rect.Contains(source) && rect.Contains(destination)) {
				if(objs[source.row,source.col] != null && objs[source.row,source.col].Remove(element)) {
					if(objs[destination.row,destination.col] == null) objs[destination.row,destination.col] = new List<T>();
					objs[destination.row,destination.col].Add(element);
					return true;
				}
			}
			return false;
		}
		public bool Add(T element, point destination) {
			// Can only add a non-null item that isn't already on the grid, and only to a point in-bounds.
			if(element != null && rect.Contains(destination) && denseObjs.Add(element)) {
				if(objs[destination.row, destination.col] == null) objs[destination.row, destination.col] = new List<T>();
				objs[destination.row, destination.col].Add(element);
				return true;
			}
			return false;
		}
		public bool Remove(T element, point source) {
			// Can only remove an existing item.
			if(element != null && rect.Contains(source)) {
				if(objs[source.row,source.col] != null && objs[source.row,source.col].Remove(element) && denseObjs.Remove(element)) {
					return true;
				}
			}
			return false;
		}
		IEnumerator IEnumerable.GetEnumerator() => this.GetEnumerator();
		public IEnumerator<T> GetEnumerator() { return denseObjs.GetEnumerator(); }
		public MultiGrid(int rows, int cols) {
			Rows = rows;
			Cols = cols;
			objs = new List<T>[rows, cols];
			denseObjs = new HashSet<T>();
		}
	}
}
