﻿using System;
using System.Collections;
using System.Collections.Generic;

namespace Points {
	public interface Positioned {
		point p { get; }
	}

	public struct point : IEquatable<point>{
		public readonly int row, col;

		public point(int row,int col) { this.row = row;  this.col = col; }
		public static point operator +(point left,point right) => new point(left.row + right.row, left.col + right.col);
		public static point operator -(point left, point right) => new point(left.row - right.row, left.col - right.col);
		public static point operator -(point right) => new point(-right.row, -right.col);
		public static point operator +(point left, int i) => new point(left.row + i, left.col + i);
		public static point operator -(point left, int i) => new point(left.row - i, left.col - i);
		public static point Zero => new point(0,0);
		public override int GetHashCode() { unchecked { return row * 7757 + col; } }
		public override bool Equals(object other) {
			if(other is point) return Equals((point)other);
			else return false;
		}
		public bool Equals(point other) => row == other.row && col == other.col;
	}

	public interface Rectangular {
		rectangle rect { get; }
	}

	public struct rectangle : IEquatable<rectangle>, IEnumerable<point>{
		public readonly point position, size;

		public rectangle(point position,point size) { this.position = position;  this.size = size; }
		IEnumerator IEnumerable.GetEnumerator() => this.GetEnumerator();
		public IEnumerator<point> GetEnumerator() {
			for(int i=position.row;i<position.row+size.row;++i) {
				for(int j= position.col;j<position.col+size.col;++j) {
					yield return new point(i,j);
				}
			}
		}
		public bool Contains(point p) => p.row >= position.row && p.row < position.row+size.row && p.col >= position.col && p.col < position.col+size.col;
		public rectangle Shrink(int i) => new rectangle(position + i, size - i);
		public rectangle Grow(int i) => new rectangle(position - i, size + i);
		public override int GetHashCode() { unchecked { return position.GetHashCode() + size.GetHashCode() * 5003; } }
		public override bool Equals(object other) {
			if(other is rectangle) return Equals((rectangle)other);
			else return false;
		}
		public bool Equals(rectangle other) => position.Equals(other.position) && size.Equals(other.size);
		//if needed, add Contains(rectangle), and Shrink(pos) & Grow(pos) which can shrink X and Y by different amounts.
	}
}
