﻿using System;
using System.Collections.Generic;

namespace StatusNamespaceTodo {
	//Maybe I'll need this one, and also a "static" version that uses Func<bool> instead of Func<bool,actor> - does that make sense?
	//I guess it makes sense, but it could always be done with some singleton-ish object.
	//
	public class StatusRuleTracker<TObject,TStatus> where TStatus : struct { //todo, decide on a name for this thing.
		private List<Func<TObject, TStatus, bool>> universalImplicationConditions;
		private List<Func<TObject, TStatus, bool>> universalSuppressionConditions;
		private Dictionary<TStatus,List<Func<TObject,bool>>> specificImplicationConditions;
		private Dictionary<TStatus,List<Func<TObject,bool>>> specificSuppressionConditions;

		public StatusRuleTracker() { } //todo
		private bool AnyTrue(IEnumerable<Func<TObject,TStatus,bool>> conditions, TObject obj, TStatus status) {
			if(conditions != null) {
				foreach(var fn in conditions) {
					if(fn(obj, status)) return true;
				}
			}
			return false;
		}
		private bool AnyTrue(IDictionary<TStatus,List<Func<TObject, bool>>> conditions, TObject obj, TStatus status) {
			if(conditions != null && conditions.ContainsKey(status)) {
				var list = conditions[status];
				if(list != null) {
					foreach(var fn in list) {
						if(fn(obj)) return true;
					}
				}
			}
			return false;
		}
		public bool HasStatus(TObject obj,TStatus status) {
			if(AnyTrue(universalImplicationConditions, obj, status) || AnyTrue(specificImplicationConditions, obj, status)) {
				if(!AnyTrue(universalSuppressionConditions, obj, status) && !AnyTrue(specificSuppressionConditions, obj, status)) {
					return true;
				}
			}
			return false;
		}
		public void AddImplication(TStatus status,Func<TObject,bool> condition) {
			if(!specificImplicationConditions.ContainsKey(status)) conditions.Add(status,new List<Func<TObject, bool>>());
			conditions[status].Add(condition);
		}
		public void RemoveImplication(TStatus status, Func<TObject, bool> condition) {
			if(conditions.ContainsKey(status)) conditions[status].Remove(condition);
		}
		public void SetBasicCondition(Func<TObject,TStatus,bool> condition) { basicCondition = condition; }
		public void AddSuppression(TStatus status, Func<TObject, bool> condition) {
			if(!suppressionConditions.ContainsKey(status)) suppressionConditions.Add(status, new List<Func<TObject, bool>>());
			suppressionConditions[status].Add(condition);
		}
		public void RemoveSuppression(TStatus status, Func<TObject, bool> condition) {
			if(suppressionConditions.ContainsKey(status)) suppressionConditions[status].Remove(condition);
		}
		public void Load(Ruleset<TObject,TStatus> ruleset) {
			if(ruleset.implications != null) {
				foreach(var pair in ruleset.implications) {
					foreach(var fn in pair.Value) {
						AddImplication(pair.Key, fn);
					}
				}
			}
			if(ruleset.suppressions != null) {
				foreach(var pair in ruleset.suppressions) {
					foreach(var fn in pair.Value) {
						AddSuppression(pair.Key, fn);
					}
				}
			}
		}
		public void Unload(Ruleset<TObject, TStatus> ruleset) {
			if(ruleset.implications != null) {
				foreach(var pair in ruleset.implications) {
					foreach(var fn in pair.Value) {
						RemoveImplication(pair.Key, fn);
					}
				}
			}
			if(ruleset.suppressions != null) {
				foreach(var pair in ruleset.suppressions) {
					foreach(var fn in pair.Value) {
						RemoveSuppression(pair.Key, fn);
					}
				}
			}
		}
		//todo, how about a helper that takes a Func<TObject,bool> and a (params) list of TStatus, so you can easily do this:
		// whatever.AddRule(a=>!Tiles[a].submerged,Attr.Blessed,Attr.Lucky,Attr.FateSmiles);
		//
		//also, *if* there is any believable way to end up with a half of a ruleset, it should be possible to add those alone.
		// (but I'm skeptical.)
	}
	public class Ruleset<TObject,TStatus> where TStatus : struct {
		public Dictionary<TStatus, List<Func<TObject, bool>>> implications; //todo, public?
		public Dictionary<TStatus, List<Func<TObject, bool>>> suppressions;
		//todo constructors and whatever
		//
		//
		//this is the one situation in which i'll put the opening brace on its own line:
		public static Ruleset<TObject,TStatus> Generate<TOtherStatus>(
			IDictionary<TOtherStatus, List<TStatus>> statusImplications,
			IDictionary<TOtherStatus, List<TStatus>> statusSuppressions,
			Func<TObject, TOtherStatus, bool> condition) where TOtherStatus : struct //todo: not sure if the value-type restriction is needed.
		{
			var result = new Ruleset<TObject,TStatus>();
			result.implications = GenerateDictionary(statusImplications,condition);
			result.suppressions = GenerateDictionary(statusSuppressions,condition);
			return result;
		}
		private static Dictionary<TStatus, List<Func<TObject, bool>>> GenerateDictionary<TOtherStatus>(
			IDictionary<TOtherStatus, List<TStatus>> statusMapping,
			Func<TObject, TOtherStatus, bool> condition) where TOtherStatus : struct
		{
			if(statusMapping == null) return null;
			var result = new Dictionary<TStatus, List<Func<TObject, bool>>>();
			foreach(var pair in statusMapping) {
				var fn = GenerateRule(condition, pair.Key);
				foreach(TStatus status in pair.Value) {
					if(!result.ContainsKey(status)) result.Add(status, new List<Func<TObject, bool>>());
					result[status].Add(fn);
				}
			}
			return result;
		}
		private static Func<TObject, bool> GenerateRule<TOtherStatus>( //todo: would a user be able to make use of this? if so, make public.
			Func<TObject, TOtherStatus, bool> condition,
			TOtherStatus otherStatus) where TOtherStatus : struct
		{
			return obj => condition(obj, otherStatus);
		}
	}
	public static class StatusDemo {
		//
		// This is a demonstration of the StatusRuleTracker class. (until I find a better name for it)
		//
		// In this demo, I assume that we're making a game where there are many important statuses to check for -
		//   such as whether an entity is flying, or on fire -
		// and that some statuses might implicitly contain other statuses.
		// (Having the UNDEAD status might imply having the NONLIVING status, for example.)
		// This is convenient if we want to define an enemy simply as UNDEAD without needing to remember what other flags it contains,
		// and equally as convenient when we want to test whether an entity is NONLIVING without needing to remember to check for UNDEAD too.
		//
		//Here are the enums we'll be using:
		public enum Attr { Flying, Levitation, Midair, Hasted, LightSensitive, CanDodge, DrainsLife, Rooted, Immobile };
			// (let's say that Levitation gives you fine control, Flying lets you go higher, and midair is just midair.)
		public enum Spell { Fireball, FlyingLeap, Levitate, SummonTarrasque };
		public enum Item { HastePotion, VampirismPotion, TeleportScroll, InfernoScroll };
		//
		//And here are the relevant parts of the Actor class:
		public class Actor {
			public HashSet<Attr> attrs;
			public HashSet<Spell> ongoingSpellEffects;
			public HashSet<Item> ongoingItemEffects;
		}
		//
		//Okay, now let's define the relationships between those enums. Some spells and some items can apply attrs.
		//
		//First, the attr->attr mappings:  (notice how easy these rules would be to load from a file)
		public static Dictionary<Attr,List<Attr>> attrRules = new Dictionary<Attr, List<Attr>> {
			[Attr.Flying] = new List<Attr> { Attr.Midair },
			[Attr.Levitation] = new List<Attr> { Attr.Midair },
		};
		//
		//Next, the spell->attr mappings:
		public static Dictionary<Spell, List<Attr>> spellRules = new Dictionary<Spell, List<Attr>> {
			[Spell.FlyingLeap] = new List<Attr> { Attr.Flying, Attr.Hasted }, //note the lack of "midair" here - it's implied!
			[Spell.Levitate] = new List<Attr> { Attr.Levitation },
		};
		//
		//and finally the item->attr mappings:
		public static Dictionary<Item, List<Attr>> itemRules = new Dictionary<Item, List<Attr>> {
			[Item.HastePotion] = new List<Attr> { Attr.Hasted, Attr.CanDodge },
			[Item.VampirismPotion] = new List<Attr> { Attr.Flying, Attr.LightSensitive, Attr.DrainsLife },
		};
		//
		//Now that all the static stuff is defined, how do we use it?
		class Game {
			//
			//Here's the StatusRuleTracker that all Actors will refer to:
			public StatusRuleTracker<Actor,Attr> Attrs = new StatusRuleTracker<Actor, Attr>();
			//
			//On game start...
			public void Initialize() {
				//
				//Note how rulesets can be generated from only those dictionaries above, and a simple lambda.
				var attrRuleset = Attrs.GenerateRuleset(attrRules, (a,attr) => Attrs.HasStatus(a,attr) ); //recursion!
				var spellRuleset = Attrs.GenerateRuleset(spellRules, (a, spell) => a.ongoingSpellEffects.Contains(spell));
				var itemRuleset = Attrs.GenerateRuleset(itemRules, (a, item) => a.ongoingItemEffects.Contains(item));
				//
				//Once generated, rulesets aren't active until loaded:
				Attrs.Load(attrRuleset);
				Attrs.Load(spellRuleset);
				Attrs.Load(itemRuleset);
				Attrs.AddImplication(Attr.CanDodge,a=>a.ongoingItemEffects == null);
				Func<Actor,bool> x = a => true;
				Attrs.AddSuppression(Attr.DrainsLife,x);
				// if we don't care about ever unloading these rules, we don't need to keep references to the rulesets.
				// (but if we do, we can then do this:)
				Attrs.Unload(attrRuleset);
			}
			//
			//Now let's see it in action:
			public void Run() {
				Actor a = new Actor();
				//pretend actual game stuff happens here
				if(!Attrs.HasStatus(a,Attr.Flying)) {
					Console.WriteLine("You fall into a bottomless pit that is inexplicably filled with rusty poisoned spikes.");
					Console.WriteLine("Game over.");
				}
			}
			//
			//That's all! For now. The final version will actually add another important concept: suppression of statuses.
			// That is, some rules can cause a certain status to be FALSE no matter how many rules say it's TRUE.
			// (Iron boots suppressing flight and levitation, for example.)
		}
	}
}
