﻿using System;
using System.Collections.Generic;

namespace StatusNamespaceTodo {
	//Maybe I'll need this one, and also a "static" version that uses Func<bool> instead of Func<bool,actor> - does that make sense?
	//I guess it makes sense, but it could always be done with some singleton-ish object.
	//
	//todo: i'm not sure about basicCondition -- could there be a need for more than 1?
	// (the basic utility of it is to provide the client the ability to, say, check a hashset with the TStatus directly,
	//    instead of filling our conditions dictionary with an entry per TStatus.)
	public class StatusRuleTracker<TObject,TStatus> where TStatus : struct { //todo, decide on a name for this thing.
		private Func<TObject,TStatus,bool> basicCondition;
		private Dictionary<TStatus,List<Func<TObject,bool>>> conditions;
		private Dictionary<TStatus,List<Func<TObject,bool>>> suppressionConditions;

		public StatusRuleTracker() { } //todo
		public bool HasStatus(TObject obj,TStatus status) {
			bool result = false;
			if(basicCondition != null && basicCondition(obj,status)) {
				result = true;
			}
			else {
				if(conditions.ContainsKey(status)) {
					var list = conditions[status];
					if(list != null) {
						foreach(var fn in list) {
							if(fn(obj)) {
								result = true;
								break;
							}
						}
					}
				}
			}
			if(result == true && suppressionConditions.ContainsKey(status)) {
				var list = suppressionConditions[status];
				if(list != null) {
					foreach(var fn in list) {
						if(fn(obj)) {
							result = false;
							break;
						}
					}
				}
			}
			return result;
		}
		public void AddCondition(TStatus status,Func<TObject,bool> condition) {
			if(!conditions.ContainsKey(status)) conditions.Add(status,new List<Func<TObject, bool>>());
			conditions[status].Add(condition);
		}
		public void RemoveCondition(TStatus status, Func<TObject, bool> condition) {
			if(conditions.ContainsKey(status)) conditions[status].Remove(condition);
		}
		public void SetBasicCondition(Func<TObject,TStatus,bool> condition) { basicCondition = condition; }
		public void AddSuppressionCondition(TStatus status, Func<TObject, bool> condition) {
			if(!suppressionConditions.ContainsKey(status)) suppressionConditions.Add(status, new List<Func<TObject, bool>>());
			suppressionConditions[status].Add(condition);
		}
		public void RemoveSuppressionCondition(TStatus status, Func<TObject, bool> condition) {
			if(suppressionConditions.ContainsKey(status)) suppressionConditions[status].Remove(condition);
		}
		public void Load(IDictionary<TStatus,List<Func<TObject,bool>>> conditions) {
			foreach(var pair in conditions) {
				foreach(var fn in pair.Value) {
					AddCondition(pair.Key,fn); //todo: or, skip the check each time (just once) and do this manually.
				}
			}
		}
		public void Unload(IDictionary<TStatus, List<Func<TObject, bool>>> conditions) {
			foreach(var pair in conditions) {
				foreach(var fn in pair.Value) {
					RemoveCondition(pair.Key, fn);
				}
			}
		}
		//todo: not sure if suppression effects should be loaded separately, or if i should change the "ruleset" objects to include a bool.
		// (or perhaps a Ruleset object contains one such dictionary for conditions, and another dictionary for suppressions.)
		public void LoadSuppression(IDictionary<TStatus, List<Func<TObject, bool>>> conditions) {
			foreach(var pair in conditions) {
				foreach(var fn in pair.Value) {
					AddSuppressionCondition(pair.Key, fn);
				}
			}
		}
		public void UnloadSuppression(IDictionary<TStatus, List<Func<TObject, bool>>> conditions) {
			foreach(var pair in conditions) {
				foreach(var fn in pair.Value) {
					RemoveSuppressionCondition(pair.Key, fn);
				}
			}
		}
		//this is the one situation in which i'll put the opening brace on its own line:
		public Dictionary<TStatus,List<Func<TObject,bool>>> GenerateRuleset<TOtherStatus>( //todo, name?
			IDictionary<TOtherStatus,List<TStatus>> statusMapping,
			Func<TObject,TOtherStatus,bool> condition) where TOtherStatus : struct //todo: not sure if the value-type restriction is needed.
		{
			var result = new Dictionary<TStatus,List<Func<TObject,bool>>>();
			foreach(var pair in statusMapping) {
				var fn = GenerateRule(condition, pair.Key);
				foreach(TStatus status in pair.Value) {
					if(!result.ContainsKey(status)) result.Add(status,new List<Func<TObject, bool>>());
					result[status].Add(fn);
				}
			}
			return result;
		}
		//todo, name?
		public Func<TObject,bool> GenerateRule<TOtherStatus>(
			Func<TObject,TOtherStatus,bool> condition,
			TOtherStatus otherStatus) where TOtherStatus : struct
		{
			return obj => condition(obj,otherStatus);
		}
	}
	public static class StatusDemo {
		//
		// This is a demonstration of the StatusRuleTracker class. (until I find a better name for it)
		//
		// In this demo, I assume that we're making a game where there are many important statuses to check for -
		//   such as whether an entity is flying, or on fire -
		// and that some statuses might implicitly contain other statuses.
		// (Having the UNDEAD status might imply having the NONLIVING status, for example.)
		// This is convenient if we want to define an enemy simply as UNDEAD without needing to remember what other flags it contains,
		// and equally as convenient when we want to test whether an entity is NONLIVING without needing to remember to check for UNDEAD too.
		//
		//Here are the enums we'll be using:
		public enum Attr { Flying, Levitation, Midair, Hasted, LightSensitive, CanDodge, DrainsLife };
			// (let's say that Levitation gives you fine control, Flying lets you go higher, and midair is just midair.)
		public enum Spell { Fireball, FlyingLeap, Levitate, SummonTarrasque };
		public enum Item { HastePotion, VampirismPotion, TeleportScroll, InfernoScroll };
		//
		//And here are the relevant parts of the Actor class:
		public class Actor {
			public HashSet<Attr> attrs;
			public HashSet<Spell> ongoingSpellEffects;
			public HashSet<Item> ongoingItemEffects;
		}
		//
		//Okay, now let's define the relationships between those enums. Some spells and some items can apply attrs.
		//
		//First, the attr->attr mappings:  (notice how easy these rules would be to load from a file)
		public static Dictionary<Attr,List<Attr>> attrRules = new Dictionary<Attr, List<Attr>> {
			[Attr.Flying] = new List<Attr> { Attr.Midair },
			[Attr.Levitation] = new List<Attr> { Attr.Midair },
		};
		//
		//Next, the spell->attr mappings:
		public static Dictionary<Spell, List<Attr>> spellRules = new Dictionary<Spell, List<Attr>> {
			[Spell.FlyingLeap] = new List<Attr> { Attr.Flying, Attr.Hasted }, //note the lack of "midair" here - it's implied!
			[Spell.Levitate] = new List<Attr> { Attr.Levitation },
		};
		//
		//and finally the item->attr mappings:
		public static Dictionary<Item, List<Attr>> itemRules = new Dictionary<Item, List<Attr>> {
			[Item.HastePotion] = new List<Attr> { Attr.Hasted, Attr.CanDodge },
			[Item.VampirismPotion] = new List<Attr> { Attr.Flying, Attr.LightSensitive, Attr.DrainsLife },
		};
		//
		//Now that all the static stuff is defined, how do we use it?
		class Game {
			//
			//Here's the StatusRuleTracker that all Actors will refer to:
			public StatusRuleTracker<Actor,Attr> Attrs = new StatusRuleTracker<Actor, Attr>();
			//
			//On game start...
			public void Initialize() {
				//
				//Note how rulesets can be generated from only those dictionaries above, and a simple lambda.
				var attrRuleset = Attrs.GenerateRuleset(attrRules, (a,attr) => Attrs.HasStatus(a,attr) ); //recursion!
				var spellRuleset = Attrs.GenerateRuleset(spellRules, (a, spell) => a.ongoingSpellEffects.Contains(spell));
				var itemRuleset = Attrs.GenerateRuleset(itemRules, (a, item) => a.ongoingItemEffects.Contains(item));
				//
				//Once generated, rulesets aren't active until loaded:
				Attrs.Load(attrRuleset);
				Attrs.Load(spellRuleset);
				Attrs.Load(itemRuleset);
				// if we don't care about ever unloading these rules, we don't need to keep references to the rulesets.
				// (but if we do, we can then do this:)
				Attrs.Unload(attrRuleset);
			}
			//
			//Now let's see it in action:
			public void Run() {
				Actor a = new Actor();
				//pretend actual game stuff happens here
				if(!Attrs.HasStatus(a,Attr.Flying)) {
					Console.WriteLine("You fall into a bottomless pit that is inexplicably filled with rusty poisoned spikes.");
					Console.WriteLine("Game over.");
				}
			}
			//
			//That's all! For now. The final version will actually add another important concept: suppression of statuses.
			// That is, some rules can cause a certain status to be FALSE no matter how many rules say it's TRUE.
			// (Iron boots suppressing flight and levitation, for example.)
		}
	}
}
